from django.db.models import Max
from django.http.response import Http404
from django.shortcuts import get_object_or_404
from django.core.exceptions import PermissionDenied
from django.conf import settings
from rest_framework.generics import (
    GenericAPIView, ListAPIView, CreateAPIView, RetrieveAPIView,
    ListCreateAPIView, RetrieveUpdateDestroyAPIView, RetrieveDestroyAPIView,
    RetrieveUpdateAPIView
)
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import ValidationError
from arkindex.project.mixins import CorpusACLMixin, SelectionMixin, DeprecatedMixin
from arkindex.project.permissions import IsVerified
from arkindex.project.openapi import AutoSchema
from arkindex.documents.models import Corpus, ElementType
from arkindex.documents.api.elements import ElementsListMixin
from arkindex.dataimport.models import \
    DataImport, DataFile, Repository, Revision, Worker, WorkerVersion, WorkerRun
from arkindex.dataimport.serializers.ml_tool import MLToolSerializer, MLToolTaskSerializer
from arkindex.dataimport.serializers.files import DataFileSerializer, DataFileCreateSerializer
from arkindex.dataimport.serializers.git import \
    RepositorySerializer, RepositoryStartImportSerializer, ExternalRepositorySerializer, RevisionSerializer
from arkindex.dataimport.serializers.imports import (
    DataImportLightSerializer, DataImportSerializer,
    DataImportFromFilesSerializer, ElementsWorkflowSerializer, WorkerRunLightSerializer, WorkerRunSerializer,
    ImportTranskribusSerializer
)
from arkindex.dataimport.serializers.workers import WorkerSerializer, WorkerVersionSerializer
from arkindex.users.models import OAuthCredentials
from arkindex.project.fields import ArrayRemove
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex_common.enums import DataImportMode
from ponos.models import State
from datetime import datetime
from uuid import UUID
import logging
import os.path

logger = logging.getLogger(__name__)


class DataImportsList(CorpusACLMixin, ListAPIView):
    """
    List all data imports
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportLightSerializer
    openapi_overrides = {
        'operationId': 'ListDataImports',
        'tags': ['imports'],
        'parameters': [
            {
                'name': 'corpus',
                'in': 'query',
                'description': 'Filter imports by corpus ID',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            },
            {
                'name': 'mode',
                'in': 'query',
                'description': 'Filter imports by mode',
                'required': False,
                'schema': {
                    'enum': [mode.value for mode in DataImportMode],
                },
            },
            {
                'name': 'id',
                'in': 'query',
                'description': 'Filter imports by beginning of UUID',
                'required': False,
                'schema': {
                    'type': 'string',
                },
            }
        ]
    }

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user)
        }

        if 'corpus' in self.request.query_params:
            corpus_id = self.request.query_params['corpus']
            try:
                corpus_id = UUID(corpus_id)
            except (AttributeError, ValueError):
                raise ValidationError({'corpus': "'{}' is not a valid UUID".format(corpus_id)})
            filters['corpus'] = self.get_corpus(corpus_id)

        if 'mode' in self.request.query_params:
            try:
                filters['mode'] = DataImportMode(self.request.query_params['mode'])
            except ValueError:
                raise ValidationError({
                    'mode': "Mode '{}' does not exist".format(self.request.query_params['mode']),
                })

        if 'id' in self.request.query_params:
            filters['id__startswith'] = self.request.query_params['id']

        qs = DataImport.objects.filter(**filters).prefetch_related('workflow__tasks')

        # Order workflow by date of last updated task in workflow
        qs = qs.annotate(last_task=Max('workflow__tasks__updated'), last_run=Max('workflow__tasks__run'))
        return qs.order_by('-last_task')


class DataImportDetails(CorpusACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve a data import
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer
    openapi_overrides = {
        'tags': ['imports'],
    }

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        if instance.state == State.Running:
            raise ValidationError("Cannot delete a workflow while it is running")
        super().perform_destroy(instance)


class DataImportRetry(CorpusACLMixin, GenericAPIView):
    """
    Retry a data import. Can only be used on imports with Error, Failed or Stopped states.
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer
    openapi_overrides = {
        'operationId': 'RetryDataImport',
        'tags': ['imports'],
    }

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def post(self, request, *args, **kwargs):
        instance = self.get_object()
        if instance.workflow is not None and instance.state in (State.Unscheduled, State.Pending):
            raise ValidationError('This workflow is already pending')
        elif instance.state == State.Running:
            raise ValidationError('This workflow is already running')
        instance.retry()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class DataImportFromFiles(CreateAPIView):
    """
    Start a data import from one or more uploaded files.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportFromFilesSerializer
    openapi_overrides = {
        'tags': ['files'],
    }

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        files = serializer.validated_data['files']
        corpus = files[0].corpus
        mode = serializer.validated_data['mode']
        # Serializer validation codes returns an element for 'folder_id'
        folder = serializer.validated_data.get('folder_id')
        folder_type = serializer.validated_data.get('folder_type')
        element_type = serializer.validated_data['element_type']

        if folder and folder.corpus != corpus:
            # The files' corpus is already validated as writable
            raise ValidationError('Element and files are in different corpora')

        if not folder and mode != DataImportMode.IIIF:
            if mode == DataImportMode.PDF:
                suffix = os.path.splitext(files[0].name)[0]
            else:
                suffix = datetime.today().strftime('%Y-%m-%d %H:%M')

            # TODO: Remove this step and let the file/IIIF import create its own folder when needed
            folder = corpus.elements.create(
                name=f'Import {mode.value} {suffix}',
                type=folder_type,
            )

        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=mode,
            element=folder,
            folder_type=folder_type,
            element_type=element_type,
        )
        self.dataimport.files.set(files)
        self.dataimport.start()


class CorpusWorkflow(SelectionMixin, CreateAPIView):
    """
    Start a Ponos ML workflow from Arkindex corpus elements
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementsWorkflowSerializer
    openapi_overrides = {
        'operationId': 'CreateElementsWorkflow',
        'description': 'Create a distributed workflow from elements of an Arkindex corpus',
        'tags': ['imports'],
        'responses': {
            '201': {
                'content': {
                    'application/json': {
                        'schema': AutoSchema()._map_serializer(DataImportSerializer())
                    }
                }
            }
        }
    }

    def create(self, request, pk=None, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        ml_tools, chunks, thumbnails, corpus, element, name, elt_type, selection, best_class, mode = map(
            lambda key: serializer.validated_data.pop(key, None),
            (
                'ml_tools',
                'chunks',
                'thumbnails',
                'corpus',
                'element',
                'name',
                'type',
                'selection',
                'best_class',
                'mode',
            )
        )

        assert mode in (DataImportMode.Elements, DataImportMode.Workers), f"DataImport mode {mode} isn't supported to create a corpus workflow"

        if selection:
            elements = self.get_selection(corpus.id)
            if not elements:
                raise ValidationError({
                    '__all__': ['No element match filtering']
                })

        element_type = None
        if elt_type:
            element_type = ElementType.objects.get(corpus=corpus, slug=elt_type)

        dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=mode,
            name_contains=name,
            element_type=element_type,
            element=element,
            best_class=best_class,
        )

        if selection:
            dataimport.elements.set(elements)

        # A Workers DataImport needs WorkerRun(s) to be set before being started
        if mode == DataImportMode.Elements:
            if not ml_tools and not thumbnails:
                raise ValidationError('Cannot start a workflow without thumbnail generation or ML tools')
            dataimport.start(
                ml_tools=MLToolTaskSerializer(ml_tools, many=True).data,
                chunks=chunks,
                thumbnails=thumbnails
            )

        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(
                dataimport,
                context=self.get_serializer_context()
            ).data,
        )


class StartProcess(APIView):
    """
    Start a process, used to build a Workflow with Workers
    """
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'StartProcess',
        'tags': ['imports'],
    }

    def post(self, request, pk=None, **kwargs):
        dataimport = get_object_or_404(DataImport, pk=self.kwargs['pk'])

        if dataimport.mode != DataImportMode.Workers or dataimport.workflow is not None:
            raise ValidationError('Only a DataImport with Workers mode and not already launched can be started later on')

        dataimport.start()

        return Response(
            status=status.HTTP_200_OK,
            data=DataImportSerializer(
                dataimport,
                context={'request': request}
            ).data,
        )


class DataImportElements(ElementsListMixin, ListAPIView):

    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListDataImportElements',
        'tags': ['imports'],
        'description': "List a DataImport's associated elements, "
                       "for DataImports running workers on existing elements.",
    })

    def get_queryset(self):
        dataimport = get_object_or_404(
            DataImport,
            pk=self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user),
        )
        return dataimport.elements.all()


class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files in a corpus
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'
    openapi_overrides = {
        'tags': ['files']
    }

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    openapi_overrides = {
        'tags': ['files'],
    }

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when a datafile is updated immediately after it's created on a small upload
        return DataFile.objects.using('default').filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        )

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        try:
            instance.s3_delete()
        except Exception as e:
            logger.warning(f'Could not delete DataFile from S3: {e}')
        return super().perform_destroy(instance)


class DataFileUpload(DeprecatedMixin, CreateAPIView):
    """
    Upload a new file to a corpus
    """
    openapi_overrides = {
        'operationId': 'UploadDataFile',
        'tags': ['files'],
    }
    deprecation_message = 'DataFile uploads via the backend are now deprecated. ' \
        'Please use `CreateDataFile`, upload to S3 using the returned `s3_put_url`, ' \
        'then use `PartialUpdateDataFile` to set its status to `checked`.'


class DataFileCreate(CreateAPIView):
    """
    Create a Datafile.
    In case of success, a signed URI is returned to upload file content directly to a remote server.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataFileCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateDataFile',
        'tags': ['files'],
    }


class GitRepositoryImportHook(APIView):
    """
    This endpoint is intended as a webhook for Git repository hosting applications like GitLab.
    """
    openapi_overrides = {
        'operationId': 'GitPushHook',
        'security': [],
        'tags': ['repos'],
    }

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        assert repo.enabled, 'No credentials available'
        repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RepositoryList(ListAPIView):
    """
    List connected repositories
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer
    openapi_overrides = {
        'operationId': 'ListRepositories',
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
        ).order_by('url')


class AvailableRepositoriesList(ListCreateAPIView):
    """
    Search for a repository to connect to.

    Using the given OAuth credentials ID, this uses the Git hosting
    application API's search feature to look for a repository matching
    the given query. Without a query, returns a full list.
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        return cred.git_provider_class(credentials=cred).list_repos(
            query=self.request.query_params.get('search'),
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])

        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(request=self.request, **serializer.validated_data)
        latest_commit_sha = provider.get_latest_commit_sha(repo)
        rev, _ = provider.get_or_create_revision(repo, latest_commit_sha)
        dataimport = rev.start_import()

        return Response(data={'import_id': str(dataimport.id)}, status=status.HTTP_201_CREATED)


class RepositoryRetrieve(CorpusACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve a repository
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(instance)


class RepositoryStartImport(RetrieveAPIView):
    """
    Start a data import on the latest revision on a repository.
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositoryStartImportSerializer
    openapi_overrides = {
        'operationId': 'StartRepositoryImport',
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user),
        )

    def retrieve(self, request, *args, **kwargs):
        repo = self.get_object()

        if not repo.enabled:
            raise ValidationError("This repository does not have any associated credentials")

        latest_commit_sha = repo.provider.get_latest_commit_sha(repo)
        rev, _ = repo.provider.get_or_create_revision(repo, latest_commit_sha)

        di = rev.start_import()

        return Response(self.get_serializer(di).data)


class RevisionRetrieve(CorpusACLMixin, RetrieveAPIView):
    """
    Retrieve a revision
    """
    permission_classes = (IsVerified, )
    serializer_class = RevisionSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Revision.objects.filter(
            repo__corpus__in=Corpus.objects.readable(self.request.user),
        )


class MLToolList(ListAPIView):
    """
    List available machine learning tools
    """
    serializer_class = MLToolSerializer
    pagination_class = None
    openapi_overrides = {
        'security': [],
        'tags': ['ml'],
    }

    def get_queryset(self):
        return sorted(
            MLTool.list(settings.ML_CLASSIFIERS_DIR),
            key=lambda tool: tool.slug,
        )


class WorkerList(ListCreateAPIView):
    """
    List workers for a given repository UUID

    Create a Worker instance on a repository for given slug, name and type, return existing
    query if a worker with the same slug already exists on the repository
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = WorkerSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Worker.objects.filter(repository_id=self.kwargs['pk'])

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        repo = get_object_or_404(Repository, pk=self.kwargs['pk'])

        worker, created = repo.workers.get_or_create(
            slug=serializer.validated_data['slug'],
            defaults={
                'name': serializer.validated_data['name'],
                'type': MLToolType(serializer.validated_data['type']),
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerSerializer(worker).data, status=reponse_status)


class WorkerVersionList(ListCreateAPIView):
    """
    List versions, their revision and associated git references for a given worker UUID

    Create a WorkerVersion instance for a worker and a revision with a given JSON configuration,
    return existing query if a workerVersion already exists for this worker and this revision
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = WorkerVersionSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return WorkerVersion.objects.filter(
            worker_id=self.kwargs['pk']
        ).prefetch_related('revision').order_by('-revision__created')

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # We need to use the default database to avoid stale read
        # when creating a worker version on a newly created worker
        try:
            worker = Worker.objects.using('default').get(pk=self.kwargs['pk'])
        except Worker.DoesNotExist:
            raise Http404

        version, created = WorkerVersion.objects.get_or_create(
            worker=worker,
            revision=serializer.validated_data['revision'],
            defaults={
                'configuration': serializer.validated_data['configuration'],
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerVersionSerializer(version).data, status=reponse_status)


class WorkerVersionRetrieve(CorpusACLMixin, RetrieveUpdateAPIView):
    """
    Retrieve a specific worker version
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerVersionSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return WorkerVersion.objects.filter(
            revision__repo__corpus__in=Corpus.objects.readable(self.request.user),
        )


class WorkerRunList(ListCreateAPIView):
    """
    List worker runs for a given dataimport UUID
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunLightSerializer
    openapi_overrides = {
        'tags': ['imports'],
    }

    def get_queryset(self):
        return WorkerRun.objects.filter(
            dataimport_id=self.kwargs['pk']
        ).order_by('id')

    def perform_create(self, serializer):
        dataimport = get_object_or_404(DataImport, pk=self.kwargs['pk'])

        if dataimport.mode != DataImportMode.Workers:
            raise ValidationError({'dataimport': ['Import mode must be Workers']})

        if dataimport.workflow is not None:
            raise ValidationError({'dataimport': ["Can't create a WorkerRun on a DataImport that has already started"]})

        serializer.save(dataimport=dataimport)


class WorkerRunDetails(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve a specific worker run
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunSerializer
    openapi_overrides = {
        'tags': ['imports'],
    }

    def get_queryset(self):
        return WorkerRun.objects.filter(
            dataimport__corpus__in=Corpus.objects.readable(self.request.user),
        )

    def perform_destroy(self, instance):
        # Remove references to this instance from its children parents field
        instance.dataimport.worker_runs.filter(parents__contains=[instance.id]).update(parents=ArrayRemove('parents', instance.id))

        return super().perform_destroy(instance)


class ImportTranskribus(CreateAPIView):
    """
    Start a data import from Transkribus email and collection ID.
    """

    permission_classes = (IsVerified, )
    serializer_class = ImportTranskribusSerializer
    openapi_overrides = {
        'operationId': 'CreateImportTranskribus',
        'description': 'Create a data import from Transkribus collection ID.',
        'tags': ['imports'],
        'responses': {
            '201': {
                'content': {
                    'application/json': {
                        'schema': AutoSchema()._map_serializer(DataImportSerializer())
                    }
                }
            }
        }
    }

    def create(self, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['transkribus']:
            raise ValidationError(['Transkribus import is unavailable due to the transkribus feature being disabled.'])
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        collection_id = serializer.validated_data['collection_id']

        # Create corpus
        corpus = Corpus.objects.create(
            name=f"Transkribus collection n°{collection_id}",
        )
        corpus.corpus_right.create(
            user=self.request.user,
            can_write=True,
            can_admin=True,
        )
        corpus.create_default_transkribus_types()

        # Create dataimport
        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=DataImportMode.Transkribus,
            collection_id=collection_id
        )
        self.dataimport.start()
