from django.urls import reverse
from rest_framework import status
from arkindex_common.enums import DataImportMode
from arkindex.dataimport.models import DataImport, DataFile
from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase
from ponos.models import State, Workflow
import uuid

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestImports(FixtureAPITestCase):
    """
    Test data imports management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.img_df = cls.corpus.files.create(
            name='test.jpg',
            size=42,
            content_type='image/jpg',
        )
        cls.pdf_df = cls.corpus.files.create(
            name='test.pdf',
            size=42,
            content_type='application/pdf',
        )
        cls.iiif_df = cls.corpus.files.create(
            name='test.json',
            size=42,
            content_type='application/json',
        )

    def setUp(self):
        super().setUp()
        self.dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Images)

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        result = data['results'][0]
        self.assertDictEqual(result, {
            'id': str(self.dataimport.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Images.value,
            'corpus': str(self.corpus.id),
            'workflow': None,
        })

    def test_list_workflow_link(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.assertIsNone(dataimport2.workflow)

        self.client.force_login(self.user)
        self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        dataimport2.refresh_from_db()

        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 2)
        results = data['results']
        self.assertListEqual(results, [{
            'id': str(self.dataimport.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Images.value,
            'corpus': str(self.corpus.id),
            'workflow': None,
        }, {
            'id': str(dataimport2.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Workers.value,
            'corpus': str(self.corpus.id),
            'workflow': f"http://testserver/ponos/v1/workflow/{dataimport2.workflow.id}/",
        }])

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        dataimport2 = corpus2.imports.create(creator=self.user, mode=DataImportMode.Images)

        response = self.client.get(reverse('api:import-list'), {'corpus': str(corpus2.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_corpus_invalid(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'corpus': 'oh-no'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'corpus': "'oh-no' is not a valid UUID"})

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        dataimport2 = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'mode': DataImportMode.PDF.value})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'mode': 'unexisting_mode'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'mode': "Mode 'unexisting_mode' does not exist"})

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first DataImport
        dataimport_id = str(uuid.uuid4())
        while dataimport_id[:10] == str(self.dataimport.id)[:10]:
            dataimport_id = str(uuid.uuid4())

        dataimport2 = self.corpus.imports.create(
            id=dataimport_id,
            creator=self.user,
            mode=DataImportMode.Images,
        )
        response = self.client.get(reverse('api:import-list'), {'id': dataimport_id[:10]})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.dataimport.workflow)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().delete()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['state'], State.Unscheduled.value)

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy_no_permission(self):
        self.client.force_login(self.user)
        acl_rights = self.corpus.corpus_right.get(user=self.user)
        acl_rights.can_write = False
        acl_rights.can_admin = False
        acl_rights.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_details_destroy_verified(self):
        self.user.verified_email = False
        self.user.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_details(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.dataimport.id))

    def test_details_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.dataimport.workflow)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().delete()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['state'], State.Unscheduled.value)

    def test_retry_requires_login(self):
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)

    def test_retry_no_workflow(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.dataimport.workflow)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        self.assertIsNotNone(self.dataimport.workflow)

    def test_retry_repo_disabled(self):
        self.client.force_login(self.user)
        self.dataimport.mode = DataImportMode.Repository
        self.dataimport.revision = self.rev
        self.dataimport.save()
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        self.creds.delete()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.creds.save()

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertListEqual(list(dataimport.files.all()), [self.img_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertEqual(dataimport.element.name, 'Import pdf test')

    def test_from_files_iiif(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.IIIF)
        self.assertListEqual(list(dataimport.files.all()), [self.iiif_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')

    def test_from_files_iiif_requires_folder_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['IIIF imports require both folder_type and element_type']
        })

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'repository',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_single(self):
        self.client.force_login(self.user)
        pdf2 = DataFile.objects.create(
            name='test2.pdf', size=1337, content_type='application/pdf', corpus=self.corpus)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id), str(pdf2.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'pdf',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_images_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_iiif_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'iiif',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name='Volume 1')
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'folder_id': str(volume.id),
                'element_type': 'page',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element, volume)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_start_process_requires_login(self):
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.dataimport.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_process_invalid_dataimport_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_start_process_wrong_dataimport_mode(self):
        self.assertNotEqual(self.dataimport.mode, DataImportMode.Workers)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.dataimport.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Only a DataImport with Workers mode and not already launched can be started later on'])

    def test_start_process_dataimport_already_started(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        dataimport2.workflow = Workflow.objects.create(recipe=RECIPE)
        dataimport2.save()
        self.assertIsNotNone(dataimport2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Only a DataImport with Workers mode and not already launched can be started later on'])

    def test_start_process(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.assertIsNone(dataimport2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport2.id))
        dataimport2.refresh_from_db()
        self.assertEqual(dataimport2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport2.workflow)
