from unittest.mock import patch, MagicMock
from django.test import override_settings
from rest_framework.reverse import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import DataImportMode
from ponos.models import State, Workflow
from arkindex.dataimport.models import DataImport, Worker, WorkerVersion
from arkindex.documents.models import Element, Corpus, DataSource
from arkindex.project.tests import FixtureAPITestCase
import yaml

RECIPE = '''
tasks:
  first:
    image: alpine
'''


class TestWorkflows(FixtureAPITestCase):
    """
    Test workflows from Arkindex existing elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume = Element.objects.get(name='Volume 1')
        cls.pages = Element.objects.get_descending(cls.volume.id).filter(type__slug='page', zone__isnull=False)
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.private_volume = cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug='volume', folder=True),
            name='Hidden',
        )
        cls.ml_tools_sample = [{'type': 'dla', 'slug': 'line-detector', 'parents': []}]

        # Workers DataImportMode
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        cls.worker_1 = Worker.objects.create(
            repository=cls.repo,
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        cls.version_1 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_1,
            configuration={"test": "test1"}
        )
        cls.version_2 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_2,
            configuration={"test": "test2"}
        )

        cls.ml_get_patch = patch('arkindex_common.ml_tool.MLTool.get')

    def setUp(self):
        super().setUp()
        self.ml_get_mock = self.ml_get_patch.start()
        dla_mock, reco_mock = MagicMock(), MagicMock()
        dla_mock.type, dla_mock.slug = MLToolType.DLAnalyser, 'line-detector'
        reco_mock.type, reco_mock.slug = MLToolType.Recognizer, 'text-recognizer'
        dla_mock.get_task_recipe.return_value = {'image': 'im1', 'command': 'cmd1'}
        reco_mock.get_task_recipe.return_value = {'image': 'im2', 'command': 'cmd2'}
        tools = {
            dla_mock.slug: dla_mock,
            reco_mock.slug: reco_mock
        }
        self.ml_get_mock.side_effect = lambda directory, type, slug: tools[slug]

    def tearDown(self):
        self.ml_get_patch.stop()
        super().tearDown()

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
        PONOS_RECIPE={'env': {'SOME_VAR': 'somevalue'}}
    )
    def test_elements_workflow(self):
        self.client.force_login(self.user)
        ml_tools = [
            {
                'type': 'dla',
                'slug': 'line-detector',
                'parents': [],
            },
            {
                'type': 'recognizer',
                'slug': 'text-recognizer',
                'parents': [{'type': 'dla', 'slug': 'line-detector'}],
            }
        ]

        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': ml_tools,
                'corpus': str(self.corpus.id),
                'type': 'page',
                'mode': 'elements'
            },
            format='json',
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertDictEqual(data, {
            'id': str(dataimport.id),
            'state': 'unscheduled',
            'corpus': str(self.volume.corpus.id),
            'files': [],
            'mode': 'elements',
            'revision': dataimport.revision,
            'workflow': reverse(
                'ponos:workflow-details',
                kwargs={'pk': workflow.id},
                request=response.wsgi_request
            ),
            'folder_type': None,
            'element_type': 'page',
            'element': None
        })
        self.assertEqual(workflow.recipes['initialisation'].image, 'tasks:latest')
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'--dataimport-id {dataimport.id} '
            f'--corpus-id {self.corpus.id} '
            '--chunks-number 1 '
            '--type page'
        )

    def test_volume_workflow(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'element': str(self.volume.id),
                'mode': 'elements'
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertEqual(dataimport.element, self.volume)
        self.assertEqual(data.pop('element')['id'], str(self.volume.id))
        self.assertDictEqual(data, {
            'id': str(dataimport.id),
            'state': 'unscheduled',
            'mode': 'elements',
            'corpus': str(self.corpus.id),
            'files': [],
            'revision': dataimport.revision,
            'workflow': reverse(
                'ponos:workflow-details',
                kwargs={'pk': workflow.id},
                request=response.wsgi_request
            ),
            'folder_type': None,
            'element_type': None,
        })

    def test_single_page_workflow(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'mode': 'elements'
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element, self.pages.first())

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
    )
    def test_chunks(self):
        ml_tools = [
            {
                'slug': 'line-detector',
                'type': 'dla',
                'parents': [],
            }, {
                'slug': 'text-recognizer',
                'type': 'recognizer',
                'parents': [{'slug': 'line-detector', 'type': 'dla'}]
            }
        ]
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': ml_tools,
                'element': str(self.volume.id),
                'corpus': str(self.corpus.id),
                'chunks': 3,
                'thumbnails': True,
                'mode': 'elements'
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertEqual(dataimport.element, self.volume)
        # 1 import, 3 thumbnails tasks, 3 times 2 machine learning tasks
        self.assertEqual(len(workflow.recipes), 10)
        self.assertListEqual(
            [(name, r.command, r.parents) for name, r in workflow.recipes.items()],
            [
                (
                    'initialisation',
                    'python -m arkindex_tasks.init_elements '
                    f'--dataimport-id {dataimport.id} '
                    f'--corpus-id {self.corpus.id} '
                    '--chunks-number 3 '
                    f'--element-id {self.volume.id}',
                    []
                ),
                ('ml_line-detector_1', 'cmd1', ['initialisation']),
                ('ml_line-detector_2', 'cmd1', ['initialisation']),
                ('ml_line-detector_3', 'cmd1', ['initialisation']),
                ('ml_text-recognizer_1', 'cmd2', ['ml_line-detector_1']),
                ('ml_text-recognizer_2', 'cmd2', ['ml_line-detector_2']),
                ('ml_text-recognizer_3', 'cmd2', ['ml_line-detector_3']),
                (
                    'thumbnails_1',
                    'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_1.json',
                    ['initialisation']
                ), (
                    'thumbnails_2',
                    'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_2.json',
                    ['initialisation']
                ), (
                    'thumbnails_3',
                    'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_3.json',
                    ['initialisation']
                )
            ]
        )

    def test_element_workflow_requires_login(self):
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'element': str(self.volume.id)
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_workflow_acl(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'element': str(self.private_volume.id),
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('element', data)
        self.assertListEqual(
            data['element'],
            ['Invalid pk "{}" - object does not exist.'.format(self.private_volume.id)]
        )

    def test_elements_name_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'name': self.pages.first().name[2:5],
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.name_contains, self.pages.first().name[2:5])

    def test_elements_type_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'type': 'page',
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element_type.slug, 'page')

    def test_elements_empty_queryset(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'type': 'volume',
                'name': 'ahahah',
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match filtering']
        )

    def test_elements_empty_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'selection': True,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match filtering']
        )

    def test_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'type': 'arachnid',
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('type', data)
        self.assertListEqual(
            data['type'],
            ['Unrecognized element type filter']
        )

    def test_mixed_element_filters(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'type': 'page',
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['Filtering parameters (name, type, best_class) cannot be used with element parameter']
        )

    def test_mixed_selection_filters(self,):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'selection': True,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['Filtering parameters (element, name, type, best_class) cannot be used with selection parameter']
        )

    def test_elements_name_escape(self,):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'name': self.pages.first().name,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'--dataimport-id {dataimport.id} '
            f'--corpus-id {self.corpus.id} '
            '--chunks-number 1 '
            f'--name-contains \'{self.pages.first().name}\'',
        )

    def test_elements_best_class_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'best_class': False,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        dataimport = DataImport.objects.get(id=response.json()['id'])
        self.assertEqual(dataimport.best_class, 'False')
        self.assertEqual(
            dataimport.workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'--dataimport-id {dataimport.id} '
            f'--corpus-id {self.corpus.id} '
            '--chunks-number 1 '
            '--best-class False',
        )

    def test_elements_best_class_id_filter(self):
        ml_class = self.corpus.ml_classes.create(name='some class')
        self.corpus.elements.first().classifications.create(
            ml_class=ml_class,
            source=DataSource.objects.first(),
            high_confidence=True,
        )
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'best_class': str(ml_class.id),
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        dataimport = DataImport.objects.get(id=response.json()['id'])
        self.assertEqual(
            dataimport.workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'--dataimport-id {dataimport.id} '
            f'--corpus-id {self.corpus.id} '
            '--chunks-number 1 '
            f'--best-class {ml_class.id}'
        )

    def test_elements_selection_filter(self):
        self.client.force_login(self.user)
        page = self.pages.first()
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [page.id]},
            format='json'
        )
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'selection': True,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow
        self.assertEqual(dataimport.elements.get(), page)
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'--dataimport-id {dataimport.id} '
            f'--corpus-id {self.corpus.id} '
            '--chunks-number 1',
        )

        response = self.client.get(reverse('api:import-elements', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(page.id))

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_elements_selection_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'corpus': str(self.corpus.id),
                'selection': True,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'selection': ['Selection is not available on this instance.']})

    def test_thumbnails_generation_only(self):
        """
        Generating thumbnails without any workflow must generate an import task
        which do not to retrieve filtered elements children
        """
        self.client.force_login(self.user)
        corpus_type = self.corpus.types.first()
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'chunks': 3,
                'type': corpus_type.slug,
                'thumbnails': True,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'--dataimport-id {dataimport.id} '
            f'--corpus-id {self.corpus.id} '
            '--chunks-number 3 '
            f'--type {corpus_type.slug} '
            '--skip-children'
        )

    def test_nothing(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'chunks': 3,
                'thumbnails': False,
                'mode': 'elements'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Cannot start a workflow without thumbnail generation or ML tools'])

    def test_workers_no_worker_runs(self):
        dataimport_2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)

        self.assertIsNone(dataimport_2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport_2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport_2.id))
        dataimport_2.refresh_from_db()
        self.assertEqual(dataimport_2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport_2.workflow)
        self.maxDiff = None
        self.assertDictEqual(yaml.safe_load(dataimport_2.workflow.recipe)['tasks'], {
            'import':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'--dataimport-id {dataimport_2.id} '
                               f'--corpus-id {self.corpus.id} '
                               '--chunks-number 1',
                    'image': 'registry.gitlab.com/arkindex/tasks'
                }
        })

    def test_workers_multiple_worker_runs(self):
        dataimport_2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        run_1 = dataimport_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )
        dataimport_2.worker_runs.create(
            version=self.version_2,
            parents=[run_1.id],
        )

        workflow_tmp = Workflow.objects.create(recipe=RECIPE)
        task_tmp = workflow_tmp.tasks.get(slug='first')
        artifact = task_tmp.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.docker_image = artifact
        self.version_1.save()
        self.version_2.docker_image = artifact
        self.version_2.save()

        self.assertIsNone(dataimport_2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport_2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport_2.id))
        dataimport_2.refresh_from_db()
        self.assertEqual(dataimport_2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport_2.workflow)
        self.assertDictEqual(yaml.safe_load(dataimport_2.workflow.recipe)['tasks'], {
            'import':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'--dataimport-id {dataimport_2.id} '
                               f'--corpus-id {self.corpus.id} '
                               '--chunks-number 1',
                    'image': 'registry.gitlab.com/arkindex/tasks'
                },
            f'worker_1_{str(self.version_1.id)[0:6]}':
                {
                    'command': None,
                    'image': f'gitlab/repo/worker_1:{self.version_1.id}',
                    'artifact': str(artifact.id),
                    'parents': ['import'],
                    'env': {'TASK_ELEMENTS': '/data/import/elements.json', 'WORKER_VERSION_ID': str(self.version_1.id)}
                },
            f'worker_1_{str(self.version_2.id)[0:6]}':
                {
                    'command': None,
                    'image': f'gitlab/repo/worker_1:{self.version_2.id}',
                    'artifact': str(artifact.id),
                    'parents': [f'worker_1_{str(self.version_1.id)[0:6]}'],
                    'env': {'TASK_ELEMENTS': '/data/import/elements.json', 'WORKER_VERSION_ID': str(self.version_2.id)}
                }
        })
