from django.urls import reverse
from django.test import override_settings
from unittest.mock import patch
from mock import AsyncMock
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.polygon import Polygon
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import Transcription, DataSource


class TestBulkTranscriptions(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.src = DataSource.objects.get(slug='test')

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_bulk_transcription(self, get_layer_mock, ml_get_mock):
        """
        Checks the view creates transcriptions, zones, links, paths and runs ES indexing
        Using bulk_transcriptions
        """
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.page.transcriptions.all().delete()
        self.client.force_login(self.internal_user)
        request_data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data=request_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.page.transcriptions.count(), 2)

        word_ts = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(word_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(word_ts.score, 0.83)
        self.assertEqual(word_ts.source, self.src)

        line_ts = self.page.transcriptions.get(text="This is a test", type=TranscriptionType.Line)
        self.assertEqual(line_ts.zone.polygon, Polygon.from_coords(0, 0, 200, 200))
        self.assertEqual(line_ts.score, 0.75)
        self.assertEqual(line_ts.source, self.src)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_bulk_transcription_no_zone(self, ml_get_mock):
        self.assertIsNone(self.vol.zone)

        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data={
            "parent": str(self.vol.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
            ]
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_update_bulk_transcription(self, get_layer_mock, ml_get_mock):
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }

        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertEqual(new_ts.source, self.src)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

        # Update again
        get_layer_mock().send.reset_mock()
        data['transcriptions'] = [
            {
                "type": "word",
                "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                "text": "NEKUDOTAYIM",
                "score": 0.83,
            },
            {
                "type": "word",
                "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                "text": "Hi",
                "score": 0.99,
            },
        ]

        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

        # Previous transcriptions should be replaced by two Word transcriptions
        self.assertEqual(self.page.transcriptions.count(), 2)

        # This transcription was inserted with the first call.
        # This ensures calling the update API with an existing transcription
        # keeps the existing transcriptions
        ts_1 = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(ts_1.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(ts_1.score, 0.83)
        self.assertEqual(ts_1.source, self.src)

        ts_2 = self.page.transcriptions.get(text="Hi", type=TranscriptionType.Word)
        self.assertEqual(ts_2.zone.polygon, Polygon.from_coords(0, 0, 200, 200))
        self.assertEqual(ts_2.score, 0.99)
        self.assertEqual(ts_2.source, self.src)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_create_bulk_transcription_no_search(self, get_layer_mock, ml_get_mock):
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        request_data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data=request_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(get_layer_mock().send.called)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_update_bulk_transcription_no_search(self, get_layer_mock, ml_get_mock):
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }

        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(get_layer_mock().send.called)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_unique_zone(self, get_layer_mock, ml_get_mock):
        """
        Checks the bulk view reuses zones when available
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = 'new_slug'
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.assertEqual(self.page.transcriptions.count(), 4)
        existing_ts = self.page.transcriptions.get(zone__image__path='img1', text="PARIS")
        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [{
                "type": "word",
                "text": "PARISH",
                "score": 0.999,
                "polygon": existing_ts.zone.polygon.serialize(),
            }]
        }
        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # A new transcription has been added with the same zone
        self.assertEqual(self.page.transcriptions.count(), 5)
        self.assertEqual(Transcription.objects.get(text="PARISH").zone.id, existing_ts.zone.id)
