from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.documents.models import \
    Element, DataSource, Transcription, Corpus
from arkindex.project.tests import FixtureAPITestCase


class TestParentsElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.private_elt = cls.private_corpus.elements.create(type=cls.private_corpus.types.create(slug='type'))
        cls.manual_source = DataSource.objects.create(type=MLToolType.Recognizer, slug='manual', internal=True)

    def setUp(self):
        self.page = self.corpus.elements.get(name='Volume 1, page 1r')

    def test_element_parents(self):
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder=false')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?type=act')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        surface = Element.objects.get(name='Surface A')
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(surface.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(surface.id)}) + '?recursive')
        self.assertEqual(len(response.json()['results']), 2)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Act 1', 'Volume 1'],
        )

        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(surface.id)})
            + '?recursive&name=aCT'
        )
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

    def test_parents_with_children_count(self):
        surface = Element.objects.get(name='Surface A')
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'recursive': True, 'with_children_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {
                'Act 1': 1,
                'Volume 1': 8,
            }
        )

    def test_with_children_count_unique(self):
        r"""
        Ensure with_children_count only counts children elements once, not even when ElementPaths are duplicated.
        Counts may have duplicates as soon as an element's parent has multiple parents:

        A   B    Element | Path
         \ /     C       | A
          C      C       | B
          |      D       | A C
          D      D       | B C

        Here, requesting the parents of D with children counts should return C with 1 child,
        although there are two ElementPaths whose last ID the path is C's.
        """
        act = Element.objects.get(name='Act 1')
        act.add_parent(Element.objects.get(name='Volume 2'))
        surface = Element.objects.get(name='Surface A')
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'with_children_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {'Act 1': 1},
        )

    def test_parent_transcription_sources_count_multiple(self):
        surface = Element.objects.get(name='Surface A')
        parent = Element.objects.get(name='Act 1')
        # Add transcriptions from two different sources on the parent
        sources = [DataSource.objects.get(slug='test'), self.manual_source]
        parent.transcriptions.bulk_create([
            Transcription(element=parent, zone=surface.zone, source=s)
            for s in sources
        ])
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'with_transcription_sources_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['transcription_sources_count'] for element in response.json()['results']},
            {
                'Act 1': 2,
            }
        )
