from django.urls import reverse
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.polygon import Polygon
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import Corpus, DataSource
from arkindex.users.models import User


class TestTranscriptions(FixtureAPITestCase):
    """
    Tests for text element view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.volume = cls.corpus.elements.get(name='Volume 1')
        cls.line = cls.corpus.elements.get(name='Text line')
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        cls.src = DataSource.objects.get(slug='test')
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()

    def test_list_transcriptions_read_right(self):
        # A read right on the element corpus is required to access transcriptions
        self.client.force_login(self.private_read_user)
        url = reverse('api:element-transcriptions', kwargs={'pk': str(self.private_page.id)})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_transcriptions_non_folder(self):
        # Transcriptions should be listed for non folder elements only
        self.client.force_login(self.user)
        url = reverse('api:element-transcriptions', kwargs={'pk': str(self.volume.id)})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Element is a folder'})

    def test_list_transcriptions_wrong_type(self):
        # Wrong transcription type
        url = reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)})
        url += '?type=potato'
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_element_transcriptions(self):
        self.page.transcriptions.all().delete()
        transcriptions = []
        for i in range(10):
            zone, _ = self.page.zone.image.zones.get_or_create(polygon=Polygon.from_coords(0, 0, i + 1, i + 1))
            # Create transcriptions on the page with their own zones
            transcriptions.append(self.page.transcriptions.create(
                source_id=self.src.id, type=TranscriptionType.Word, zone=zone,
            ))
        self.client.force_login(self.user)

        with self.assertNumQueries(10):
            response = self.client.get(reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()['results']
        self.assertEqual(len(results), 10)
        self.assertCountEqual(
            [(tr['id'], tr['element']) for tr in results],
            # Element should not be serialized in case recursive parameter is not set
            [(str(tr.id), None) for tr in transcriptions]
        )

    def test_list_transcriptions_recursive(self):
        for i in range(1, 5):
            # Add 4 transcriptions on the page line
            self.line.transcriptions.create(source_id=self.src.id, type=TranscriptionType.Line, text=f'Text {i}')
        for i in range(1, 5):
            # Add 4 transcribed line children
            zone, _ = self.page.zone.image.zones.get_or_create(polygon=Polygon.from_coords(0, 0, i + 1, i + 1))
            line = self.page.corpus.elements.create(zone=zone, type=self.line.type, name=f'Added line {i}')
            line.transcriptions.create(source_id=self.src.id, type=TranscriptionType.Line, text=f'Added text {i}')
            line.add_parent(self.page)

        self.client.force_login(self.user)

        with self.assertNumQueries(14):
            response = self.client.get(
                reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)}),
                data={'recursive': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()['results']
        self.assertEqual(len(results), 12)

        page_polygon = [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]]
        line_polygon = [[400, 400], [400, 500], [500, 500], [500, 400], [400, 400]]
        self.assertCountEqual(
            [(data['element']['type'], data['element']['zone']['polygon'], data['text']) for data in results],
            [
                ('page', page_polygon, 'PARIS'),
                ('page', page_polygon, 'ROY'),
                ('page', page_polygon, 'Lorem ipsum dolor sit amet'),
                ('page', page_polygon, 'DATUM'),
                ('text_line', line_polygon, 'Text 1'),
                ('text_line', line_polygon, 'Text 2'),
                ('text_line', line_polygon, 'Text 3'),
                ('text_line', line_polygon, 'Text 4'),
                ('text_line', [[0, 0], [0, 2], [2, 2], [2, 0], [0, 0]], 'Added text 1'),
                ('text_line', [[0, 0], [0, 3], [3, 3], [3, 0], [0, 0]], 'Added text 2'),
                ('text_line', [[0, 0], [0, 4], [4, 4], [4, 0], [0, 0]], 'Added text 3'),
                ('text_line', [[0, 0], [0, 5], [5, 5], [5, 0], [0, 0]], 'Added text 4')
            ]
        )

    def test_list_transcriptions_recursive_filtered(self):
        for i in range(1, 5):
            # Add 4 transcriptions on the page line
            self.line.transcriptions.create(source_id=self.src.id, type=TranscriptionType.Line, text=f'Text {i}')

        self.client.force_login(self.user)

        with self.assertNumQueries(12):
            response = self.client.get(
                reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)}),
                data={'recursive': 'true', 'type': 'line'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()['results']
        self.assertEqual(len(results), 4)
        for tr in results:
            self.assertEqual(tr.get('type'), 'line')
