from collections import namedtuple
from django.db import connection
from arkindex_common.enums import TranscriptionType
from arkindex.project.polygon import Polygon
from arkindex.documents.models import Element
from arkindex.images.models import Zone
import csv
import io
import uuid

# Link a transcription data with a Polygon
# This is hashable (box is hashable)
TrPolygon = namedtuple('TrPolygon', 'image_id, element_id, source_id, type, polygon, text, score')


def build_transcriptions(image, parent, source, items):
    """
    Build needed TrPolygon namedtuples from parsed index files
    """
    assert isinstance(parent, Element)

    # Build all TrPolygon from items
    required = {
        TrPolygon(
            image.id,
            parent.id,
            source.id,
            i.get('type', TranscriptionType.Word),
            Polygon.from_dict(i),
            i['text'],
            float(i['score']),
        )
        for i in items
    }

    # Build all TrPolygon from existing
    existing = {
        TrPolygon(
            # A transcription zone may be defined by its parent (null value)
            tr.zone and tr.zone.image_id,
            tr.element_id,
            tr.source_id,
            tr.type,
            tr.zone and tr.zone.polygon,
            tr.text,
            tr.score,
        )
        for tr in parent.transcriptions.using('default').all().prefetch_related('zone')
    }

    # Calc needed TrPolygon to build
    return list(required.difference(existing))


def save_transcriptions(*tr_polygons):
    """
    Save TrPolygons in the database
    Uses PostgreSQL's COPY, as it is the fastest way to import, by generating CSV data in file-like objects
    """

    assert len(tr_polygons) > 0, \
        'No polygons, so no transcriptions to save.'

    zone_ids = {
        (image_id, polygon): zone_id
        for zone_id, image_id, polygon
        in Zone.objects
               .filter(image_id__in={tr_polygon.image_id for tr_polygon in tr_polygons})
               .values_list('id', 'image_id', 'polygon')
    }

    new_zones, transcriptions = [], []
    for tr_polygon in tr_polygons:
        key = (tr_polygon.image_id, tr_polygon.polygon)
        zone_id = zone_ids.get(key)
        if not zone_id:
            zone_id = uuid.uuid4()
            zone_ids[key] = zone_id
            new_zones.append(Zone(
                id=zone_id,
                image_id=tr_polygon.image_id,
                polygon=tr_polygon.polygon,
            ))

        transcriptions.append((
            uuid.uuid4(),
            tr_polygon.element_id,
            tr_polygon.source_id,
            zone_id,
            tr_polygon.type.value,
            tr_polygon.text,
            tr_polygon.score,
        ))

    Zone.objects.bulk_create(new_zones)

    copy_data = io.StringIO()
    writer = csv.writer(copy_data, dialect='unix', quoting=csv.QUOTE_ALL)
    writer.writerows(transcriptions)
    copy_data.seek(0)

    # Convert to CSV and run COPY via psycopg2
    with connection.cursor() as cursor:
        cursor.copy_expert(
            """
            COPY documents_transcription (id, element_id, source_id, zone_id, type, text, score) FROM STDIN
            WITH (FORMAT csv, DELIMITER ',', QUOTE '"')
            """,
            copy_data,
        )

    return transcriptions
