from pathlib import Path
from rest_framework.schemas.openapi import SchemaGenerator as BaseSchemaGenerator
import yaml
from django.conf import settings

PATCH_FILE = Path(__file__).absolute().parent / 'patch.yml'


class SchemaGenerator(BaseSchemaGenerator):
    """
    Custom schema generator that applies the custom patches from patch.yml.

    Django REST Framework allows generating schemas depending on the authentication;
    if a user is not an admin for example, the permissions will be checked, and some
    endpoints may be ignored.  While this can be useful, it makes schema generation-
    specific code in views or serializers, for example for default values, much more
    complex as there are more specific cases to gracefully handle, and could require
    re-downloading the schema in the API client on each login/logout.
    This can also help with caching as the generated schema simply never changes.

    Therefore, to avoid this situation, we ask DRF to generate "public" schemas all
    the time, as if we were running manage.py generateschema.
    """

    def get_schema(self, *args, **kwargs):
        # Always send public=True and no request
        schema = super().get_schema(request=None, public=False)

        with PATCH_FILE.open() as f:
            patch = yaml.safe_load(f)

        self.patch_paths(schema['paths'], patch.pop('paths', {}))
        schema.update(patch)
        schema['info']['version'] = settings.VERSION

        return schema

    def patch_paths(self, paths, patches):
        for path_name, methods in patches.items():
            assert path_name in paths, f'OpenAPI path {path_name} does not exist'
            path_object = paths[path_name]

            for method, operation_patch in methods.items():
                assert method in path_object, f'Method {method} on OpenAPI path {path_name} does not exist'
                operation_object = path_object[method]

                # Update responses separately to allow adding status codes without replacing all of them
                if 'responses' in operation_patch and 'responses' in operation_object:
                    operation_object['responses'].update(operation_patch.pop('responses'))

                operation_object.update(operation_patch)
