from django.conf import settings
from rest_framework import serializers
from enum import Enum
from uuid import UUID
from urllib.parse import quote, unquote
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex.project.polygon import Point, Polygon


class EnumField(serializers.ChoiceField):
    """
    Serializes an enum field into a JSON value
    """

    def __init__(self, enum, *args, **kwargs):
        assert issubclass(enum, Enum) or not enum
        self.enum = enum
        choices = [(item.value, item.name) for item in self.enum]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        if not isinstance(obj, self.enum):
            obj = self.to_internal_value(obj)
        return obj.value

    def to_internal_value(self, data):
        assert self.enum is not None, "No enum set on EnumField"
        try:
            return self.enum(data)
        except ValueError:
            raise serializers.ValidationError('Value is not of type {}'.format(self.enum.__name__))


class DataSourceSlugField(serializers.CharField):

    def __init__(self, tool_type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if tool_type is not None:
            assert isinstance(tool_type, MLToolType)
        self.tool_type = tool_type

    def to_representation(self, obj):
        from arkindex.documents.models import DataSource
        if isinstance(obj, DataSource):
            return super().to_representation(obj.slug)
        return super().to_representation(obj)

    def to_internal_value(self, data):
        from arkindex.documents.models import DataSource

        slug = super().to_internal_value(data)
        if self.tool_type:
            if slug == 'manual':
                # Get or create a manual source for this tool.
                # This behavior is not RESTful but allows to serialize
                # a manual source for any tool.
                manual_source, _ = DataSource.objects.get_or_create(
                    type=self.tool_type,
                    slug=slug,
                    defaults={
                        'revision': '',
                        'internal': False,
                    }
                )
                return manual_source

            # Try to get or create the source from an ML tool,
            # and fall back to just using any existing source with this type
            try:
                ml_tool = MLTool.get(settings.ML_CLASSIFIERS_DIR, self.tool_type, slug)
            except ValueError:
                pass
            else:
                return DataSource.from_ml_tool(ml_tool)

        # Pick the most recent source by ordering by revision
        # Note that there still might be duplicate sources if there is no tool type filter
        # as DataSource only requires unique slugs by type and revision;
        # there could be a "sklearn 0.1" classifier and a "sklearn 0.1" DLA.
        queryset = DataSource.objects.filter(slug=slug).order_by('-revision')
        if self.tool_type:
            queryset = queryset.filter(type=self.tool_type)

        source = queryset.first()
        if not source:
            raise serializers.ValidationError('Source with slug {!r} not found'.format(data))
        return source


class PointField(serializers.ListField):
    child = serializers.IntegerField()
    min_length = 2
    max_length = 2

    def to_representation(self, point):
        return [point.x, point.y]

    def to_internal_value(self, coords):
        try:
            return Point(*super().to_internal_value(coords))
        except (AssertionError, TypeError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class PolygonField(serializers.ListField):
    child = PointField()
    min_length = 3

    def to_internal_value(self, data):
        try:
            return Polygon(super().to_internal_value(data))
        except (AssertionError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class IIIFPathField(serializers.CharField):
    """
    An IIIF image path field where special characters are escaped,
    including slashes.
    """

    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        # If unquoting the value gives a different value, the value was already quoted
        if '/' in value or unquote(value) == value:
            value = quote(value, safe='')
        return value


class BestClassField(serializers.UUIDField):

    def to_internal_value(self, data):
        try:
            return UUID(data)
        except (TypeError, ValueError, AttributeError):
            return str(data).lower() not in ('false', '0')
